# Copyright 2023 Deepgram SDK contributors. All Rights Reserved.
# Use of this source code is governed by a MIT license that can be found in the LICENSE file.
# SPDX-License-Identifier: MIT

from importlib import import_module
import logging, verboselogs
from typing import Union

from ..options import DeepgramClientOptions
from .errors import DeepgramModuleError

# live client
# classes and input
from .prerecorded import (
    PreRecordedClient,
    AsyncPreRecordedClient,
    PrerecordedOptions,
)

# responses
from .prerecorded import (
    AsyncPrerecordedResponse,
    PrerecordedResponse,
    SyncPrerecordedResponse,
)

# live client
# classes and input
from .live import (
    LiveClient,
    AsyncLiveClient,
    LiveOptions,
    LiveTranscriptionEvents,
)

# responses
from .live import (
    LiveResultResponse,
    MetadataResponse,
    SpeechStartedResponse,
    UtteranceEndResponse,
    ErrorResponse,
)


class Listen:
    """
    Represents a client for interacting with the Deepgram API.

    This class provides a client for making requests to the Deepgram API with various configuration options.

    Attributes:
        api_key (str): The Deepgram API key used for authentication.
        config_options (DeepgramClientOptions): An optional configuration object specifying client options.

    Raises:
        DeepgramApiKeyError: If the API key is missing or invalid.

    Methods:
        live: (Preferred) Returns a Threaded LiveClient instance for interacting with Deepgram's transcription services.
        prerecorded: (Preferred) Returns an Threaded PreRecordedClient instance for interacting with Deepgram's prerecorded transcription services.

        asynclive: Returns an (Async) LiveClient instance for interacting with Deepgram's transcription services.
        asyncprerecorded: Returns an (Async) PreRecordedClient instance for interacting with Deepgram's prerecorded transcription services.
    """

    def __init__(self, config: DeepgramClientOptions):
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(logging.StreamHandler())
        self.logger.setLevel(config.verbose)
        self.config = config

    @property
    def prerecorded(self) -> PreRecordedClient:
        return self.Version(self.config, "prerecorded")

    @property
    def asyncprerecorded(self) -> AsyncPreRecordedClient:
        return self.Version(self.config, "asyncprerecorded")

    @property
    def live(self) -> LiveClient:
        return self.Version(self.config, "live")

    @property
    def asynclive(self) -> AsyncLiveClient:
        return self.Version(self.config, "asynclive")

    # INTERNAL CLASSES
    class Version:
        def __init__(self, config, parent: str):
            self.logger = logging.getLogger(__name__)
            self.logger.addHandler(logging.StreamHandler())
            self.logger.setLevel(config.verbose)
            self.config = config
            self.parent = parent

        # FUTURE VERSIONING:
        # When v2 or v1.1beta1 or etc. This allows easy access to the latest version of the API.
        # @property
        # def latest(self):
        #     match self.parent:
        #         case "live":
        #             return LiveClient(self.config)
        #         case "prerecorded":
        #             return PreRecordedClient(self.config)
        #         case _:
        #             raise DeepgramModuleError("Invalid parent")

        def v(
            self, version: str = ""
        ) -> Union[
            LiveClient, AsyncLiveClient, PreRecordedClient, AsyncPreRecordedClient
        ]:
            self.logger.debug("Version.v ENTER")
            self.logger.info("version: %s", version)
            if len(version) == 0:
                self.logger.error("version is empty")
                self.logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Invalid module version")

            parent = ""
            fileName = ""
            className = ""
            match self.parent:
                case "live":
                    parent = "live"
                    fileName = "client"
                    className = "LiveClient"
                case "asynclive":
                    parent = "live"
                    fileName = "async_client"
                    className = "AsyncLiveClient"
                case "prerecorded":
                    parent = "prerecorded"
                    fileName = "client"
                    className = "PreRecordedClient"
                case "asyncprerecorded":
                    parent = "prerecorded"
                    fileName = "async_client"
                    className = "AsyncPreRecordedClient"
                case _:
                    self.logger.error("parent unknown: %s", self.parent)
                    self.logger.debug("Version.v LEAVE")
                    raise DeepgramModuleError("Invalid parent type")

            # create class path
            path = f"deepgram.clients.{parent}.v{version}.{fileName}"
            self.logger.info("path: %s", path)
            self.logger.info("className: %s", className)

            # import class
            mod = import_module(path)
            if mod is None:
                self.logger.error("module path is None")
                self.logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Unable to find package")

            my_class = getattr(mod, className)
            if my_class is None:
                self.logger.error("my_class is None")
                self.logger.debug("Version.v LEAVE")
                raise DeepgramModuleError("Unable to find class")

            # instantiate class
            myClass = my_class(self.config)
            self.logger.notice("Version.v succeeded")
            self.logger.debug("Version.v LEAVE")
            return myClass
